import nanoid
from pathlib import Path
from roughrider.storage.fs import FilesystemStorage


class FlatStorage(FilesystemStorage):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id_size = kwargs.get('id_size', 16)

    def generate_ticket(self) -> str:
        return nanoid.generate(
            '_-23456789abcdefghijkmnpqrstuvwxyzABCDEFGHIJKMNPQRSTUVWXYZ',
            size=self.id_size)

    def ticket_to_uri(self, uid: str) -> Path:
        return self.root / uid



import pytest
from roughrider.storage.meta import StorageCenter, FileInfo
from roughrider.storage.flat import FlatStorage


def test_empty_store(test_file):
    center = StorageCenter()
    with pytest.raises(LookupError) as exc:
        center.store('somewhere', test_file)
    assert str(exc.value) == 'Namespace `somewhere` is unknown.'


def test_empty_retrieve():
    center = StorageCenter()
    with pytest.raises(LookupError) as exc:
        center.retrieve('somewhere', 'bogus_id')
    assert str(exc.value) == 'Namespace `somewhere` is unknown.'


def test_empty_get():
    center = StorageCenter()
    info = FileInfo(
        namespace='somewhere',
        ticket='12345678-1234-5678-1234-56781234567a',
        size=28,
        checksum=('md5', '53195454e1210adae36ecb34453a1f5a'),
        metadata={}
    )
    with pytest.raises(LookupError) as exc:
        center[info]
    assert str(exc.value) == 'Namespace `somewhere` is unknown.'


def test_register(tmp_path):
    center = StorageCenter()
    flat = FlatStorage('somewhere', tmp_path)
    center.register(flat)
    assert 'somewhere' in center.namespaces

    someother = FlatStorage('somewhere', tmp_path)
    with pytest.raises(NameError) as exc:
        center.register(someother)
    assert str(exc.value) == 'Namespace `somewhere` already exists.'


def test_store_get_retrieve_delete(test_file, tmp_path):
    center = StorageCenter()
    flat = FlatStorage('somewhere', tmp_path)
    center.register(flat)

    info = center.store('somewhere', test_file)
    assert isinstance(info, dict)

    test_file.seek(0)
    fiter = center[info]
    assert b''.join(fiter) == test_file.read()

    test_file.seek(0)
    fiter = center.retrieve('somewhere', info['ticket'])
    assert b''.join(fiter) == test_file.read()

    del center[info]
    with pytest.raises(FileNotFoundError):
        center.retrieve('somewhere', info['ticket'])



import enum
import hashlib
from abc import ABC, abstractmethod
from functools import partial
from pathlib import Path
from typing import TypedDict, BinaryIO, MutableMapping, Iterable, Tuple


ChecksumAlgorithm = enum.Enum(
    'Algorithm', {
        name: getattr(hashlib, name)
        for name in hashlib.algorithms_guaranteed
    }
)


class FileInfo(TypedDict):
    ticket: str
    size: int
    checksum: tuple[str, str]  # (algorithm, value)
    namespace: str
    metadata: dict | None = None


class Storage(ABC):
    name: str
    root: Path

    @abstractmethod
    def generate_ticket(self) -> str:
        pass

    @abstractmethod
    def ticket_to_uri(self, uid: str) -> Path:
        pass

    @abstractmethod
    def retrieve(self, ticket: str) -> Iterable[bytes]:
        pass

    @abstractmethod
    def store(self, data: BinaryIO, **metadata) -> FileInfo:
        pass

    @abstractmethod
    def delete(self, ticket: str) -> bool:
        pass


class StorageCenter:

    __slots__ = ('namespaces',)

    namespaces: MutableMapping[str, Storage]

    def __init__(self, namespaces=None):
        if namespaces is None:
            namespaces = {}
        self.namespaces = namespaces

    def __getitem__(self, info: FileInfo) -> Iterable[bytes]:
        return self.retrieve(info['namespace'], info['ticket'])

    def __delitem__(self, info: FileInfo):
        return self.delete(info['namespace'], info['ticket'])

    def register(self, storage: Storage):
        if storage.name in self.namespaces:
            raise NameError(f'Namespace `{storage.name}` already exists.')
        self.namespaces[storage.name] = storage

    def store(self, namespace: str, data: BinaryIO, **metadata) -> FileInfo:
        storage = self.namespaces.get(namespace)
        if storage is None:
            raise LookupError(f'Namespace `{namespace}` is unknown.')
        return storage.store(data, **metadata)

    def retrieve(self, namespace: str, ticket: str) -> Iterable[bytes]:
        storage = self.namespaces.get(namespace)
        if storage is None:
            raise LookupError(f'Namespace `{namespace}` is unknown.')
        return storage.retrieve(ticket)

    def delete(self, namespace: str, ticket: str) -> bool:
        storage = self.namespaces.get(namespace)
        if storage is None:
            raise LookupError(f'Namespace `{namespace}` is unknown.')
        return storage.delete(ticket)



import pytest
import re
from typing import Iterator
from roughrider.storage.meta import FileInfo
from roughrider.storage.flat import FlatStorage
from unittest import mock


def mock_nanoid(*args, **kwargs):
    return 'my_tiny_id'


@mock.patch('nanoid.generate', mock_nanoid)
def test_ticket(tmp_path):
    storage = FlatStorage('flat', tmp_path)
    ticket = storage.generate_ticket()
    assert ticket == 'my_tiny_id'
    path = storage.ticket_to_uri(ticket)
    assert path == tmp_path / ticket


def test_ticket_format(tmp_path):
    storage = FlatStorage('flat', tmp_path)
    ticket = storage.generate_ticket()
    assert len(ticket) == 16
    assert re.match(r'[^\w\-]', ticket) is None

    storage.id_size = 10
    ticket = storage.generate_ticket()
    assert len(ticket) == 10
    assert re.match(r'[^\w\-]', ticket) is None


@mock.patch('nanoid.generate', mock_nanoid)
def test_store(test_file, tmp_path):
    storage = FlatStorage('flat', tmp_path)
    storage_info = storage.store(test_file)
    assert storage_info == FileInfo(
        namespace='flat',
        ticket='my_tiny_id',
        size=28,
        checksum=('md5', '53195454e1210adae36ecb34453a1f5a'),
        metadata={}
    )


def test_retrieve(test_file, tmp_path):
    storage = FlatStorage('flat', tmp_path)
    storage_info = storage.store(test_file)
    iterator = storage.retrieve(storage_info['ticket'])
    assert isinstance(iterator, Iterator)
    test_file.seek(0)
    assert b''.join(iterator) == test_file.read()


def test_delete(test_file, tmp_path):
    storage = FlatStorage('flat', tmp_path)
    storage_info = storage.store(test_file)
    storage.delete(storage_info['ticket'])

    with pytest.raises(FileNotFoundError):
        storage.delete(storage_info['ticket'])


def test_checksum(test_file, tmp_path):
    storage = FlatStorage('flat', tmp_path, algorithm="sha256")
    storage_info = storage.store(test_file)
    assert storage_info['checksum'] == (
        'sha256',
        '18e9b7c9c1be46b1c62938b11b02f513a4d507630c4aee744799df83e0a94ba6'
    )

    with pytest.raises(LookupError) as exc:
        FlatStorage('flat', tmp_path, algorithm="pouet")
    assert str(exc.value) == "Unknown algorithm: `pouet`."



from typing import Iterable, BinaryIO
from pathlib import Path
from roughrider.storage.meta import FileInfo, Storage, ChecksumAlgorithm


class FilesystemStorage(Storage):

    checksum_algorithm: ChecksumAlgorithm

    def __init__(self, name: str, root: Path, algorithm='md5'):
        self.name = name
        self.root = root
        try:
            self.checksum_algorithm = ChecksumAlgorithm[algorithm]
        except KeyError:
            raise LookupError(f'Unknown algorithm: `{algorithm}`.')

    @staticmethod
    def file_iterator(path: Path, chunk=4096):
        with path.open('rb') as reader:
            while True:
                data = reader.read(chunk)
                if not data:
                    break
                yield data

    def retrieve(self, ticket: str) -> Iterable[bytes]:
        path = self.ticket_to_uri(ticket)
        if not path.exists():
            raise FileNotFoundError(path)
        return self.file_iterator(path)

    def delete(self, ticket: str) -> Iterable[bytes]:
        path = self.ticket_to_uri(ticket)
        try:
            path.unlink()
            return True
        except FileNotFoundError:
            raise  # we need to propagate.
        return False

    def store(self, data: BinaryIO, **metadata) -> FileInfo:
        ticket = self.generate_ticket()
        path = self.ticket_to_uri(ticket)
        assert not path.exists()  # this happens on ticket conflicts.
        depth = len(path.relative_to(self.root).parents)
        if depth > 1:
            path.parent.mkdir(mode=0o755, parents=True, exist_ok=False)
        size = 0
        fhash = self.checksum_algorithm.value()
        with path.open('wb+') as target:
            for block in iter(lambda: data.read(4096), b""):
                size += target.write(block)
                fhash.update(block)

        return FileInfo(
            namespace=self.name,
            ticket=ticket,
            size=size,
            checksum=(fhash.name, fhash.hexdigest()),
            metadata=metadata
        )



import uuid
import pytest
from pathlib import Path
from typing import Iterator


def pyfs_installed():
    try:
        import fs  # noqa F401
        return True
    except ImportError:
        return False


@pytest.mark.skipif(not pyfs_installed(), reason="PyFilesystem needed.")
def test_bushy_pyfs(test_file):
    from fs.memoryfs import MemoryFS
    from fs.tempfs import TempFS
    from fs.errors import ResourceNotFound
    from roughrider.storage.pyfs import PyFSStorage
    from roughrider.storage.meta import FileInfo


    class BushyStorage(PyFSStorage):

        count = None

        def generate_ticket(self) -> str:
            if self.count is None:
                self.count = 0
            self.count += 1
            return str(uuid.UUID(int=self.count))

        def ticket_to_uri(self, uid: str) -> Path:
            return Path(f'{uid[0:4]}/{uid[4:8]}/{uid[9:]}')


    storage = BushyStorage('bushy', fs=MemoryFS())
    storage_info = storage.store(test_file)
    assert storage_info == FileInfo(
        namespace='bushy',
        ticket='00000000-0000-0000-0000-000000000001',
        size=28,
        checksum=('md5', '53195454e1210adae36ecb34453a1f5a'),
        metadata={}
    )
    iterator = storage.retrieve(storage_info['ticket'])
    assert isinstance(iterator, Iterator)
    test_file.seek(0)
    assert b''.join(iterator) == test_file.read()
    test_file.seek(0)

    storage.delete(storage_info['ticket'])

    with pytest.raises(ResourceNotFound):
        storage.delete(storage_info['ticket'])


    storage = BushyStorage('bushy', fs=TempFS())
    storage_info = storage.store(test_file)
    assert storage_info == FileInfo(
        namespace='bushy',
        ticket='00000000-0000-0000-0000-000000000001',
        size=28,
        checksum=('md5', '53195454e1210adae36ecb34453a1f5a'),
        metadata={}
    )
    iterator = storage.retrieve(storage_info['ticket'])
    assert isinstance(iterator, Iterator)
    test_file.seek(0)
    assert b''.join(iterator) == test_file.read()
    test_file.seek(0)

    storage.delete(storage_info['ticket'])

    with pytest.raises(ResourceNotFound):
        storage.delete(storage_info['ticket'])



try:
    from typing import Iterable, BinaryIO, Type, ClassVar
    from pathlib import Path
    from fs.base import FS
    from roughrider.storage.meta import FileInfo, Storage, ChecksumAlgorithm


    class PyFSStorage(Storage):

        fs: FS
        checksum_algorithm: ChecksumAlgorithm

        def __init__(self, name: str, fs: FS, algorithm='md5'):
            self.name = name
            self.fs = fs
            try:
                self.checksum_algorithm = ChecksumAlgorithm[algorithm]
            except KeyError:
                raise LookupError(f'Unknown algorithm: `{algorithm}`.')

        def retrieve(self, ticket: str) -> Iterable[bytes]:
            path = self.ticket_to_uri(ticket)
            if not self.fs.exists(str(path)):
                raise FileNotFoundError(path)

            def file_iterator(path: Path, chunk=4096):
                with self.fs.openbin(str(path)) as reader:
                    while True:
                        data = reader.read(chunk)
                        if not data:
                            break
                        yield data

            return file_iterator(path)

        def delete(self, ticket: str) -> Iterable[bytes]:
            path = self.ticket_to_uri(ticket)
            try:
                self.fs.remove(str(path))
                return True
            except FileNotFoundError:
                raise  # we need to propagate.
            return False

        def store(self, data: BinaryIO, **metadata) -> FileInfo:
            ticket = self.generate_ticket()
            path = self.ticket_to_uri(ticket)
            self.fs.makedirs(str(path.parent), recreate=True)
            size = 0
            fhash = self.checksum_algorithm.value()
            with self.fs.openbin(str(path), mode='w+') as target:
                for block in iter(lambda: data.read(4096), b""):
                    size += target.write(block)
                    fhash.update(block)

            return FileInfo(
                namespace=self.name,
                ticket=ticket,
                size=size,
                checksum=(fhash.name, fhash.hexdigest()),
                metadata=metadata
            )
except ImportError:
    pass



import pytest
import uuid
from unittest import mock
from typing import Iterator
from roughrider.storage.meta import FileInfo
from roughrider.storage.bushy import BushyStorage


def mock_uuid():
    return uuid.UUID(int=0x12345678123456781234567812345678)


@mock.patch('uuid.uuid1', mock_uuid)
def test_ticket(tmp_path):
    storage = BushyStorage('bushy', tmp_path)
    ticket = storage.generate_ticket()
    assert ticket == '12345678-1234-5678-1234-567812345678'

    path = storage.ticket_to_uri(ticket)
    assert path == (
        tmp_path / '1234' / '5678' / '1234-5678-1234-567812345678')

    with pytest.raises(ValueError) as exc:
        storage.ticket_to_uri('random ticket')
    assert str(exc.value) == 'Invalid ticket format.'


@mock.patch('uuid.uuid1', mock_uuid)
def test_store(test_file, tmp_path):
    storage = BushyStorage('bushy', tmp_path)
    storage_info = storage.store(test_file)
    assert storage_info == FileInfo(
        namespace='bushy',
        ticket='12345678-1234-5678-1234-567812345678',
        size=28,
        checksum=('md5', '53195454e1210adae36ecb34453a1f5a'),
        metadata={}
    )


@mock.patch('uuid.uuid1', mock_uuid)
def test_store_metadata(test_file, tmp_path):
    storage = BushyStorage('bushy', tmp_path)
    storage_info = storage.store(
        test_file, filename="test.jpg", owner="admin")
    assert storage_info == FileInfo(
        namespace='bushy',
        ticket='12345678-1234-5678-1234-567812345678',
        size=28,
        checksum=('md5', '53195454e1210adae36ecb34453a1f5a'),
        metadata={'filename': 'test.jpg', 'owner': 'admin'}
    )


def test_retrieve(test_file, tmp_path):
    storage = BushyStorage('bushy', tmp_path)
    storage_info = storage.store(test_file)
    iterator = storage.retrieve(storage_info['ticket'])
    assert isinstance(iterator, Iterator)
    test_file.seek(0)
    assert b''.join(iterator) == test_file.read()


def test_delete(test_file, tmp_path):
    storage = BushyStorage('bushy', tmp_path)
    storage_info = storage.store(test_file)
    storage.delete(storage_info['ticket'])

    with pytest.raises(FileNotFoundError):
        storage.delete(storage_info['ticket'])


def test_checksum(test_file, tmp_path):
    storage = BushyStorage('bushy', tmp_path, algorithm="sha256")
    storage_info = storage.store(test_file)
    assert storage_info['checksum'] == (
        'sha256',
        '18e9b7c9c1be46b1c62938b11b02f513a4d507630c4aee744799df83e0a94ba6'
    )
    with pytest.raises(LookupError) as exc:
        BushyStorage('bushy', tmp_path, algorithm="pouet")
    assert str(exc.value) == "Unknown algorithm: `pouet`."



import re
import uuid
from pathlib import Path
from roughrider.storage.fs import FilesystemStorage


UUID = re.compile(
    "^[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}$")


class BushyStorage(FilesystemStorage):

    def generate_ticket(self) -> str:
        return str(uuid.uuid1())

    def ticket_to_uri(self, uid: str) -> Path:
        if not UUID.match(uid):
            raise ValueError('Invalid ticket format.')
        return self.root / uid[0:4] / uid[4:8] / uid[9:]
